function Om = error_AWGN_PPV_OnX_LB(n, R, Pe, kappa)
%
% Om = error_AWGN_PPV_OnX_LB(n, R, Pe, kappa)
%
% inputs :
%  n    : packet length
%  R    : code rate
%  Pe   : reference packet error probability
% kappa : O(n^-kappa) approximation. when kappa=2 or 3 compact closed form
%         results are used, when kappa>3 a serie is evaluated where only 
%         the best terms are kept (see [1]). Note: the method is slow for
%         kappa>3. The suggested value (default) is kappa=2.
%
% the first three must be vectors of the same length, and kappa a scalar
%
% outputs:
%  Om    : O(n^-kappa) approximation of the SNR needed to match the PPV
%          meta-converse bound on error probability
%
% The method builds upon the O(n^-k) approximation of the PPV bound from:
% [1] T. Erseghe, "On the Evaluation of the Polyanskiy-Poor-Verdu Converse 
% Bound for Finite Block-Length Coding in AWGN", 
% http://arxiv.org/abs/1401.7169
% [2] T. Erseghe, "Coding in the Finite-Blocklength Regime: Bounds based 
% on Laplace Integrals and their Asymptotic Approximations", 
% http://arxiv.org/abs/1511.04629
%
% (c) T. Erseghe 2016

if ~exist('kappa')
    kappa = 2; % default value
end

options1 = optimset('Algorithm','interior-point',...
    'GradObj','on',...
    'GradCon','on',...
    'Display', 'off',...
    'TolFun',1e-20,...
    'TolX', 1e-20,...
    'MaxIter',1000);
options2 = optimset('Algorithm','interior-point',...
    'GradObj','on',...
    'GradCon','off',...
    'Display', 'off',...
    'TolFun',1e-20,...
    'TolX', 1e-20,...
    'MaxIter',1000);

for k = 1:length(n)
    if (n(k)==inf)
        Om(k) = 2^(2*R(k))-1;
    else
        nn = n(k);
        ra = R(k);
        pe = Pe(k);
        % 1) find couple (lambda,om)
        Qe = fmincon(@(x)zero_obj(x),0,[],[],[],[],[],[],...
                     @(x)constr_Qe(x,pe),options1);
        om0 = 1.2*(exp(2*R)-1);
        V = om0.*(2+om0)/2./((1+om0).^2);
        if (kappa<=3)
            % starting point from (132) in the Proof of Theo.10 in [2]
            la0 = 1 + sqrt(4*V/nn)*Qe;
        else
            % see [1]
            la0 = log(1+om0)/2-Qe*sqrt((om0^3)/2/nn/(1+om0)^2/(2+om0));
        end
        x = fmincon(@(x)zero_obj(x),[la0;om0],[],[],[],[],[],[],...
                     @(x)constr_om(x,ra,pe,nn,kappa),options2);
        % 2) extract the SNR value
        Om(k) = x(2);
    end
end


% ------------------- FUNCTIONS FOR OPTIMIZATION -------------------------

function y = log_FA(la, Om, n, kappa)
if (kappa<=3)&&((la<=1)||(la>=1+2*Om)) % exit if out of bounds
    y = nan;
    return
elseif (kappa>3)&&((la<=log(1+Om/(1+Om))/2)||(la>=log(1+Om)/2)) 
    y = nan;
    return
end
if (kappa==2) % O(n^-2) approx in [2]
    r0 = (sqrt(Om^2+4*la*(1+Om))-Om)/2/(1+Om);
    r0 = 1/r0;
    sa = (r0-1)/2/Om; % saddle point
    al = (1+Om)/Om*(1/r0-1)-log(r0)+2*la*sa;
    a2 = (-2*Om/r0).^2.*((1+Om)/Om/r0+1/2);
    y = 0.5*al-0.5*log(2*pi*n*a2*sa^2)/n;
elseif (kappa==3) % O(n^-3) approx in [2]
    r0 = (sqrt(Om^2+4*la*(1+Om))-Om)/2/(1+Om);
    r0 = 1/r0;
    sa = (r0-1)/2/Om; % saddle point
    al = (1+Om)/Om*(1/r0-1)-log(r0)+2*la*sa;
    a2 = (-2*Om/r0).^2.*((1+Om)/Om/r0+1/2);
    a3 = (-2*Om/r0).^3.*((1+Om)/Om/r0+1/3);
    a4 = (-2*Om/r0).^4.*((1+Om)/Om/r0+1/4);
    g = (12*a2*a4-15*a3^2)/(8*a2^3)-(3*a3)/(2*a2^2*sa)-1/(a2*sa^2);
    y = 0.5*al-0.5*log(2*pi*n*a2*sa^2)/n+log(1+g/n)/n;
else % best out of kappa coefficients, see [1]
    [y pre] = coefficients(la,Om,n,kappa,1);    
end

function y = log_MD(la, Om, n, kappa)
if (kappa<=3)&&((la<=1)||(la>=1+2*Om)) % exit if out of bounds
    y = nan;
    return
elseif (kappa>3)&&((la<=log(1+Om/(1+Om))/2)||(la>=log(1+Om)/2)) 
    y = nan;
    return
end
if (kappa==2) % O(n^-2) approx in [2]
    r0 = (sqrt(Om^2+4*la*(1+Om))-Om)/2;
    r0 = 1/r0;
    sa = (r0-1)*(1+Om)/2/Om; % saddle point
    al = (1/r0-1)/Om-log(r0)+2*la*sa;
    a2 = (-2*Om/(1+Om)/r0)^2.*(1/Om/r0+1/2);
    y = 0.5*al-0.5*log(2*pi*n*a2*sa^2)/n;
elseif (kappa==3) % O(n^-3) approx in [2]
    r0 = (sqrt(Om^2+4*la*(1+Om))-Om)/2;
    r0 = 1/r0;
    sa = (r0-1)*(1+Om)/2/Om; % saddle point
    al = (1/r0-1)/Om-log(r0)+2*la*sa;
    a2 = (-2*Om/(1+Om)/r0)^2.*(1/Om/r0+1/2);
    a3 = (-2*Om/(1+Om)/r0)^3.*(1/Om/r0+1/3);
    a4 = (-2*Om/(1+Om)/r0)^4.*(1/Om/r0+1/4);
    g = (12*a2*a4-15*a3^2)/(8*a2^3)-(3*a3)/(2*a2^2*sa)-1/(a2*sa^2);
    y = 0.5*al-0.5*log(2*pi*n*a2*sa^2)/n+log(1+g/n)/n;
else % best out of kappa coefficients see [1]
    [y pre] = coefficients(la,Om,n,kappa,0);
end

function [out pre] = coefficients(ga,Om,n,Nc,isFA)
Nc = 2;
N = 2*Nc-1;
sg = sinh(ga);
cg = cosh(ga);
th = log(Om)-log(2*sg)-(isFA>0)*log(1+Om);
s  = sign(ga-th);
a = @(x) cg-cosh(x)+sg*(x-ga);
v = -a(th)/sg;
u0 = s*sqrt(2*v*sg);
% derive Taylor coefficients ck
nuk = [0 0];
nuk(3:2:N+2) = -cg./gamma(3:2:N+2);
nuk(4:2:N+2) = -sg./gamma(4:2:N+2);
Q = eye(N+1);
for j = 2:N+1
    for k = j+1:N+1
        l = 1:k-j;
        cc = ((l*j-k+j+l)/(k-j)).*(((-1).^l)./(l+1));
        Q(j,k) = cc*Q(j,k-l).';
    end
end
ak = [0 0];
for k = 2:N+1
    ak(k+1) = (nuk(3:k+1)*Q(2:k,k))*exp(-k*ga);
end
bk = [0, -sqrt(2*ak(3))];
for k = 2:N
    bk(k+1) = (ak(k+2) - 0.5*sum(bk(3:k).*bk(k:-1:3)))/bk(2);
end
P = diag(bk(2).^(1:N));
for j = 1:N
    for k = j+1:N
        l = 1:k-j;
        cc = ((l*j-k+j+l)/(k-j)).*(bk(l+2)/bk(2));
        P(j,k) = cc*P(j,k-l).';
    end
end
tk = [exp(ga), 1/bk(2)];
for k = 2:N
    tk(k+1) = -(tk(2:k)*P(1:k-1,k))/(bk(2)^k);
end
S = diag(tk(2).^(1:N));
for j = 1:N
    for k = j+1:N
        l = 1:k-j;
        cc = ((l*j-k+j+l)/(k-j)).*(tk(l+2)/tk(2));
        S(j,k) = cc*S(j,k-l).';
    end
end
xk = [1/(exp(th)-exp(ga))];
for k = 1:N
    xk(k+1) = 1./((exp(th)-exp(ga)).^(2:k+1))*S(1:k,k);
end
for k = 0:N-1
    fk(k+1) = ((1:k+1).*tk(2:k+2))*(xk(k+1:-1:1)).';
end
ck = fk(1:2:end)/1i;   
% check correctness of first factor
if abs(1/sqrt(cg)/(exp(th-ga)-1)-ck(1))/abs(ck(1))>1e-8 
    disp('help!!!');
end
g = (ck/(2*sqrt(pi))) ...
    .* (gamma(0.5:1:Nc-0.5)/gamma(0.5)) ...
    .* ((4*sg/n).^(0.5:1:Nc-0.5));
tg = cumsum(g);
[~,pos] = min(abs(g));
tg = tg(pos(1));
pre = abs(g(pos(1)))/abs(tg); % precision
out = -0.5*v+log(abs(tg))/n;
if ((2*isFA-1)*s<0)
    out = log(1-exp(n*out))/n;
end
    
function [f gradf] = zero_obj(x)
f = [0];
gradf = [0];

function [c ceq gradc gradceq] = constr_Qe(x,Pe)
c = [];
gradc = [];
ceq = [0.5*erfc((x)/sqrt(2))-Pe];
gradceq = [-exp(-x.^2/2)/sqrt(2*pi)];

function [c ceq gradc gradceq] = constr_om(x,R,Pe,n,kappa)
c = [];
gradc = [];
ceq = [log_MD(x(1),x(2),n,kappa)-log(Pe)/n; ...
       log_FA(x(1),x(2),n,kappa)+R*log(2)];
gradceq = []; % too long to derive !!!

