function R = rate_AWGN_RCU_On2_LB(n, Om, Pe)
%
% R = rate_AWGN_RCU_On2_LB(n, Om, Pe)
%
% inputs :
%  n    : packet length
%  Om   : SNR value
%  Pe   : reference packet error probability
%
% must be vectors of the same length 
%
% outputs:
%  R    : O(n^-2) approximation of the RCU bound on rate
%
% The method builds upon the O(n^-2) approximation of the RCU bound from:
% T. Erseghe, "Coding in the Finite-Blocklength Regime: Bounds based on 
% Laplace Integrals and their Asymptotic Approximations", 
% http://arxiv.org/abs/1511.04629
%
% (c) T. Erseghe 2016

options1 = optimset('Algorithm','interior-point',...
    'GradObj','on',...
    'GradCon','on',...
    'Display', 'off',...
    'TolFun',1e-20,...
    'TolX', 1e-20,...
    'MaxIter',1000);
options2 = optimset('Algorithm','interior-point',...
    'GradObj','on',...
    'GradCon','off',...
    'Display', 'off',...
    'TolFun',1e-20,...
    'TolX', 1e-20,...
    'MaxIter',1000);

for k = 1:length(n)
    if (n(k)==inf)
        R(k) = log2(1+Om(k))/2;
    else
        nn = n(k);
        om = Om(k);
        al = sqrt(om/4);
        pe = Pe(k);
        % 1) find lambda
        Qe = fmincon(@(x)zero_obj(x),0,[],[],[],[],[],[],...
                     @(x)constr_Qe(x,pe),options1);
        % starting point from (155) in the Proof of Theo.18
        la0 = sqrt(om/(1+om))-sqrt((2+om)/2/(1+om)^3/nn)*Qe; 
        la = fmincon(@(x)zero_obj(x),la0,[],[],[],[],[],[],...
                     @(x)constr_Pe(x,nn,al,pe),options2);
        % 2) find saddle point sa
        sa0 = -2*la/(1-la^2); % starting point from paragraph before (68)
        sa = fmincon(@(x)zero_obj(x),sa0,[],[],[],[],[],[],...
                     @(x)constr_sa(x,nn,la),options1);
        % 3) evaluate R
        R(k) = -(G0(sa,nn/2)+la*sa)/2/log(2)+log2(pi*nn*sa^2*G2(sa,nn/2))/2/nn;
        if (sa>0)||(R(k)<0)||(~isreal(R(k)))
            R(k) = nan;
        end
    end
end



% ------------------- FUNCTIONS FOR OPTIMIZATION -------------------------

function y = log_Pe(a,n,al)
aal = a*al;
aal2 = aal.^2;
w1 = aal+sqrt(1+aal2);
u0 = 0.5*log(1-a.^2)-2*al^2+aal2+aal*sqrt(1+aal2)+log(w1);
w0 = (2*a-2*al*(1-a.^2).*w1).*(2*al*(1-a.^2).*w1-a).*sqrt((1-a.^2).*(1+aal.*w1)./a.^2);
y = u0-log(2*pi*n)/2/n-log(w0)/n;

function y = G0(s,nu)
x = sqrt(1+s.^2);
if (nu==inf)
    y = x-1-log((1+x)/2);
else
    y = x-1-log(x)/2/nu-(1-1/nu)*log((1+x)/2);
end

function y = G1(s,nu)
x = sqrt(1+s.^2);
g1 = 1-1./x/2/nu-(1-1/nu)./(1+x);
y = g1.*s./x;

function y = G2(s,nu)
x = sqrt(1+s.^2);
g1 = 1-1./x/2/nu-(1-1/nu)./(1+x);
g2 = 1./x.^2/2/nu+(1-1/nu)./(1+x).^2;
y = g2.*(s./x).^2+g1./x.^3;

function [f gradf] = zero_obj(x)
f = [0];
gradf = [0];

function [c ceq gradc gradceq] = constr_Qe(x,Pe)
c = [];
gradc = [];
ceq = [0.5*erfc((x)/sqrt(2))-Pe];
gradceq = [-exp(-x.^2/2)/sqrt(2*pi)];

function [c ceq gradc gradceq] = constr_Pe(x,nn,al,Pe)
c = [];
gradc = [];
ceq = [log_Pe(x,nn,al)-log(Pe)/nn];
gradceq = [];

function [c ceq gradc gradceq] = constr_sa(x,nn,la)
c = [];
gradc = [];
ceq = [G1(x,nn/2)+la];
gradceq = [G2(x,nn/2)];
