function Om = error_BIAWGN_NA_LB(n, R, Pe)
%
% Om = error_BIAWGN_NA_LB(n, R, Pe)
%
% inputs :
%  n    : packet length
%  R    : code rate
%  Pe   : reference packet error probability
%
% these must be vectors of the same length
%
% outputs:
%  Om    : SNR needed to match the normal approximation bound on error 
%          probability
%
% The method used builds upon the normal approximation, e.g., see 
% T. Erseghe, "Coding in the Finite-Blocklength Regime: Bounds based on 
% Laplace Integrals and their Asymptotic Approximations", 
% http://arxiv.org/abs/1511.04629
%
% (c) T. Erseghe 2016


q = @(x) 0.5*erfc((x)/sqrt(2));
options = optimset('TolX',1e-10,'TolFun',1e-10);
options2 = optimset('Algorithm','interior-point',...
    'GradObj','on',...
    'GradCon','off',...
    'Display', 'off',...
    'TolFun',1e-20,...
    'TolX', 1e-20,...
    'MaxIter',1000);
om0 = 1;

for k = 1:length(n)
    nn = n(k);
    ra = R(k);
    pe = Pe(k);        
    Qe = fminsearch(@(x) abs(q(x)-pe),0,options);
    x = fmincon(@(x)zero_obj(x),om0,[],[],[],[],[],[],...
                 @(x)constr_om(x,ra,Qe,nn),options2);     
    Om(k) = x;
    om0 = x;
    disp(['Pe = ' num2str(pe) ', n = ' num2str(nn) ...
          ', om = ' num2str(10*log10(Om(k))) 'dB'])
end

% ------------------- FUNCTIONS FOR OPTIMIZATION -------------------------

function y = rate(om,nn,Qe)

h = @(x) log(1+exp(-2*x));
fHp = @(x,Om) exp(-(x-Om).^2/2/Om)/sqrt(2*pi*Om);
I = min(40*sqrt(om)+50*om+2,200);
tol = 1e-12;

Hp1 = quadl(@(x)fHp(x,om).*(-h(x)),-I,I,tol);
C = 1 + log2(exp(1))*Hp1;
if (nn==inf)
    y = C;
else
    Hp2 = quadl(@(x)fHp(x,om).*h(x).^2,-I,I,tol);
    V  = Hp2-Hp1^2;
    y = C-Qe*log2(exp(1))*sqrt(V/nn)+log2(nn)/2/nn;
end

function [f gradf] = zero_obj(x)
f = [0];
gradf = [0];

function [c ceq gradc gradceq] = constr_om(x,ra,Qe,nn)
c = [];
gradc = [];
ceq = [rate(x,nn,Qe)-ra];
gradceq = []; % too long to derive !!!



    
    