function R = rate_BIAWGN_NA_UB(n, Om, Pe)
%
% R = rate_BIAWGN_NA_UB(n, Om, Pe)
%
% inputs :
%  n    : packet length
%  Om   : SNR value
%  Pe   : reference packet error probability
%
% must be vectors of the same length 
%
% outputs:
%  R    : normal approximation bound on rate
%
% The method used builds upon the normal approximation, e.g., see 
% T. Erseghe, "Coding in the Finite-Blocklength Regime: Bounds based on 
% Laplace Integrals and their Asymptotic Approximations", 
% http://arxiv.org/abs/1511.04629
%
% (c) T. Erseghe 2016

% functions definition
q = @(x) 0.5*erfc((x)/sqrt(2));
h = @(x) log(1+exp(-2*x));
fHp = @(x,Om) exp(-(x-Om).^2/2/Om)/sqrt(2*pi*Om);
I = @(Om) min(40*sqrt(Om)+50*Om+2,200);
tol = 1e-12;

for k = 1:length(n)
    nn = n(k);
    om = Om(k);
    pe = Pe(k);
    Hp1 = quadl(@(x)fHp(x,om).*(-h(x)),-I(om),I(om),tol);
    C = 1 + log2(exp(1))*Hp1;
    if (nn==inf)
        R(k) = C;
    else
        Hp2 = quadl(@(x)fHp(x,om).*h(x).^2,-I(om),I(om),tol);
        V  = Hp2-Hp1^2;
        Qe = fminsearch(@(x) abs(q(x)-pe),0,optimset('TolX',1e-10,'TolFun',1e-10));
        R(k) = C-Qe*log2(exp(1))*sqrt(V/nn)+log2(nn)/2/nn;
        if (R(k)<0)||(~isreal(R(k)))
            R(k) = nan;
        end
    end
end

