function R = rate_BIAWGN_PPV_On2_UB(n, Om, Pe)
%
% R = rate_BIAWGN_PPV_On2_UB(n, Om, Pe)
%
% inputs :
%  n    : packet length
%  Om   : SNR value
%  Pe   : reference packet error probability
%
% the first three must be vectors of the same length, and kappa a scalar
%
% outputs:
%  R    : O(n^-2) approximation of the PPV meta-converse bound on rate
%
% The method builds upon the O(n^-2) approximation of the PPV bound from:
% T. Erseghe, "Coding in the Finite-Blocklength Regime: Bounds based 
% on Laplace Integrals and their Asymptotic Approximations", 
% http://arxiv.org/abs/1511.04629
%
% (c) T. Erseghe 2016


q = @(x) 0.5*erfc((x)/sqrt(2));
h = @(x) log(1+exp(-2*x));
fHp = @(x,Om) exp(-(x-Om).^2/2/Om)/sqrt(2*pi*Om);
tol = 1e-12;
options = optimset('TolX',tol,'TolFun',tol);
options2 = optimset('Algorithm','interior-point',...
    'GradObj','on',...
    'GradCon','off',...
    'Display', 'off',...
    'TolFun',1e-20,...
    'TolX', 1e-20,...
    'MaxIter',1000);

for k = 1:length(n)
    nn = n(k);
    om = Om(k);
    pe = Pe(k);
    % find saddle point
    if nn==inf
        sb = 0;
    else
        if (k==1)
            Qe = fminsearch(@(x) abs(q(x)-pe),0,options);
            I = min(40*sqrt(om)+50*om+2,200);
            Hp1 = quadl(@(x)fHp(x,om).*(-h(x)),-I,I,tol);
            Hp2 = quadl(@(x)fHp(x,om).*h(x).^2,-I,I,tol);
            V  = Hp2-Hp1^2;
            sb0 = -Qe/sqrt(V*nn);
        end
        sb = fmincon(@(x)zero_obj(x),sb0,[],[],[],[],[],[],...
                 @(x)constr_om(x,om,pe,nn),options2);     
        sb0 = sb;
    end
    % find rate
    R(k) = -log_FA(sb,om,nn)/log(2);
    if (R(k)<0)||(~isreal(R(k)))
        R(k) = nan;
    end
    disp(['Om = ' num2str(10*log10(om)) 'dB, R = ' num2str(R(k))])
end


% ------------------- FUNCTIONS FOR OPTIMIZATION -------------------------

function y = log_FA(sb,Om,n)

h = @(x) log(1+exp(-2*x));
fHp = @(x) exp(-(x-Om).^2/2/Om-sb*h(x))/sqrt(2*pi*Om);
I = min(40*sqrt(Om)+50*Om+2,200);
tol = 1e-12;

Hp  = quadl(@(x)fHp(x),-I,I,tol);
Hp1 = quadl(@(x)fHp(x).*(-h(x)),-I,I,tol);
la = -Hp1./Hp;
sa = sb+1;
if n==inf
    y = la*sa+log(Hp/2);
else
    Hp2 = quadl(@(x)fHp(x).*h(x).^2,-I,I,tol);
    a2  = Hp2./Hp-(Hp1./Hp).^2;
    y = la*sa+log(Hp/2)-log(2*pi*n*sa^2*a2)/2/n;
    if(0) % extension for O(n^-3) approximation
        Hp3 = quadl(@(x)fHp(x).*(-h(x)).^3,-I,I,tol);
        Hp4 = quadl(@(x)fHp(x).*h(x).^4,-I,I,tol);
        a3  = (Hp3./Hp-3*Hp2.*Hp1./(Hp).^2+2*(Hp1./Hp).^3)/3;
        a4  = (Hp4./Hp-4*Hp3.*Hp1./(Hp).^2-3*(Hp2./Hp).^2 ...
                +12*Hp2.*(Hp1)^2./(Hp).^3-6*(Hp1./Hp).^4)/12;
        g = (12*a2.*a4-15*a3.^2)./a2.^3/8 - 3*a3./(2*sa.*a2.^2)-1./(a2.*sa.^2);    
        y = y + log(1+g/n)/n;
    end
end

function y = log_MD(sb,Om,n)

h = @(x) log(1+exp(-2*x));
fHp = @(x) exp(-(x-Om).^2/2/Om-sb*h(x))/sqrt(2*pi*Om);
I = min(40*sqrt(Om)+50*Om+2,200);
tol = 1e-12;

Hp  = quadl(@(x)fHp(x),-I,I,tol);
Hp1 = quadl(@(x)fHp(x).*(-h(x)),-I,I,tol);
la = -Hp1./Hp;
if n==inf
    y = la.*sb+log(Hp);
else
    Hp2 = quadl(@(x)fHp(x).*h(x).^2,-I,I,tol);
    a2  = Hp2./Hp-(Hp1./Hp).^2;
    y = la.*sb+log(Hp) -log(2*pi*n*sb.^2.*a2)/2/n;
    if(0) % extension for O(n^-3) approximation
        Hp3 = quadl(@(x)fHp(x).*(-h(x)).^3,-I,I,tol);
        Hp4 = quadl(@(x)fHp(x).*h(x).^4,-I,I,tol);
        a3  = (Hp3./Hp-3*Hp2.*Hp1./(Hp).^2+2*(Hp1./Hp).^3)/3;
        a4  = (Hp4./Hp-4*Hp3.*Hp1./(Hp).^2-3*(Hp2./Hp).^2 ...
                +12*Hp2.*(Hp1)^2./(Hp).^3-6*(Hp1./Hp).^4)/12;
        g = (12*a2.*a4-15*a3.^2)./a2.^3/8 - 3*a3./(2*sb.*a2.^2)-1./(a2.*sb.^2);    
        y = y + log(1+g/n)/n;
    end
end

function [f gradf] = zero_obj(x)
f = [0];
gradf = [0];

function [c ceq gradc gradceq] = constr_om(s,om,pe,nn)
c = [];
gradc = [];
ceq = [log_MD(s,om,nn)-log(pe)/nn];
gradceq = []; % too long to derive !!!



