function Om = error_BSC_NA_LB(n, R, Pe)
%
% Om = error_BSC_NA_LB(n, R, Pe)
%
% inputs :
%  n    : packet length
%  R    : code rate
%  Pe   : reference packet error probability
%
% these must be vectors of the same length
%
% outputs:
%  Om    : SNR needed to match the normal approximation bound on error 
%          probability
%
% The method used builds upon the normal approximation, e.g., see 
% T. Erseghe, "Coding in the Finite-Blocklength Regime: Bounds based on 
% Laplace Integrals and their Asymptotic Approximations", 
% http://arxiv.org/abs/1511.04629
%
% (c) T. Erseghe 2016


q = @(x) 0.5*erfc((x)/sqrt(2));
h = @(x) -x.*log2(x)-(1-x).*log2(1-x);
options = optimset('TolX',1e-10,'TolFun',1e-10);

for k = 1:length(n)
    if (n(k)==inf)
        C = @(Pb) 1-h(Pb);
        pb = @(om) q(sqrt(om));
        Ra = @(x) C(pb(x));
        Om(k) = fminsearch(@(x) abs(Ra(x)-ra),1,options);
    else
        pe = Pe(k);
        ra = R(k);
        Qe = fminsearch(@(x) abs(q(x)-pe),0,options);
        V = @(Pb) Pb*(1-Pb)*log((1-Pb)/Pb)^2;
        C = @(Pb) 1-h(Pb);
        pb = @(om) q(sqrt(om));
        Ra = @(x) C(pb(x))-sqrt(V(pb(x))/n(k))*log2(exp(1))*Qe+log2(n(k))/2/n(k);
        Om(k) = fminsearch(@(x) abs(Ra(x)-ra),1,options);
    end
end

