function Om = error_BSC_RCU_On2_UB(n, R, Pe)
%
% Om = error_BSC_RCU_On2_UB(n, R, Pe)
%
% inputs :
%  n    : packet length
%  R    : code rate
%  Pe   : reference packet error probability
%
% these must be vectors of the same length
%
% outputs:
%  Om    : O(n^-2) approximation of the SNR needed to match the RCU
%          achievability bound on error probability
%
% The method builds upon the O(n^-k) approximation of the PPV bound from:
% T. Erseghe, "Coding in the Finite-Blocklength Regime: Bounds based 
% on Laplace Integrals and their Asymptotic Approximations", 
% http://arxiv.org/abs/1511.04629
%
% (c) T. Erseghe 2016

options1 = optimset('Algorithm','interior-point',...
    'GradObj','on',...
    'GradCon','on',...
    'Display', 'off',...
    'TolFun',1e-20,...
    'TolX', 1e-20,...
    'MaxIter',1000);
options2 = optimset('Algorithm','interior-point',...
    'GradObj','on',...
    'GradCon','off',...
    'Display', 'off',...
    'TolFun',1e-20,...
    'TolX', 1e-20,...
    'MaxIter',1000);

for k = 1:length(n)
    if (n(k)==inf)
        disp('inf n not supported')
        Om(k) = nan;
    else
        nn = n(k);
        ra = R(k);
        pe = Pe(k);
        % 1) find x = [lambda,sa,om]
        Qe = fmincon(@(x)zero_obj(x),0,[],[],[],[],[],[],...
                     @(x)constr_Qe(x,pe),options1);
        % starting point from (155) in the Proof of Theo.18
        om0 = 1;
        Pb0 = 0.5*erfc(sqrt(om0/2));
        la0 = min(1.011*Pb0,1-1e-8); % this works ok
        x = fmincon(@(x)zero_obj(x),[la0;om0],[],[],[],[],[],[],...
                     @(x)constr_om(x,nn,ra,pe),options2);
        % 2) extract SNR value
        Om(k) = x(2);
    end
end



% ------------------- FUNCTIONS FOR OPTIMIZATION -------------------------

function y = rate(s, n)
he = -s.*log(s)-(1-s).*log(1-s);
w2 = log(2)-he+log(2*pi*n*s.*(1-2*s).^2./(1-s))/2/n;
y = w2;

function y = log_Pe(s, Pb, n)
he = -s.*log(s)-(1-s).*log(1-s);
w0 = he+log(1-Pb)-s*log((1-Pb)/Pb)-log(2*pi*n*s.*(1-s))/2/n;
w1 = Pb*(1-s)./(s-Pb)+Pb*(1-s).^2./(Pb*(1-s).^2-s.^2*(1-Pb));
if(w1>0)
    y = w0+log(w1)/n;
else 
    y = nan;
end

function [f gradf] = zero_obj(x)
f = [0];
gradf = [0];

function [c ceq gradc gradceq] = constr_Qe(x,Pe)
c = [];
gradc = [];
ceq = [0.5*erfc((x)/sqrt(2))-Pe];
gradceq = [-exp(-x.^2/2)/sqrt(2*pi)];

function [c ceq gradc gradceq] = constr_om(x,nn,R,Pe)
c = [];
gradc = [];
ceq = [log_Pe(x(1),0.5*erfc(sqrt(x(2)/2)),nn)-log(Pe)/nn;...
       rate(x(1),nn)-R*log(2)...
       ];
gradceq = [];

