function R = rate_BSC_NA_UB(n, Om, Pe)
%
% R = rate_BSC_NA_UB(n, Om, Pe)
%
% inputs :
%  n    : packet length
%  Om   : SNR value
%  Pe   : reference packet error probability
%
% must be vectors of the same length 
%
% outputs:
%  R    : normal approximation bound on rate
%
% The method used builds upon the normal approximation, e.g., see 
% T. Erseghe, "Coding in the Finite-Blocklength Regime: Bounds based on 
% Laplace Integrals and their Asymptotic Approximations", 
% http://arxiv.org/abs/1511.04629
%
% (c) T. Erseghe 2016


q = @(x) 0.5*erfc((x)/sqrt(2));
h = @(x) -x.*log2(x)-(1-x).*log2(1-x);
options = optimset('TolFun',1e-12,'MaxFunEvals',10000,'MaxIter',10000);

for k = 1:length(n)
    om = Om(k);
    nn = n(k);
    pe = Pe(k);
    Pb = q(sqrt(om));
    C = 1-h(Pb);
    if isinf(nn)
        R(k) = C;
    else
        V = Pb*(1-Pb)*log((1-Pb)/Pb)^2;
        Qe = fminsearch(@(x)abs(q(x)-pe),1,options);
        R(k) = C-sqrt(V/nn)/log(2)*Qe+log2(2*pi*nn)/2/nn;
        if (R(k)<0)||(~isreal(R(k)))
            R(k) = nan;
        end
    end
end

