function R = rate_BSC_PPV_OnX_UB(n, Om, Pe, kappa)
%
% R = rate_BSC_PPV_OnX_UB(n, Om, Pe, kappa)
%
% inputs :
%  n    : packet length
%  Om   : SNR value
%  Pe   : reference packet error probability
% kappa : O(n^-kappa) approximation. when kappa=2 or 3 compact closed form
%         results are used, when kappa>3 a serie is evaluated where only 
%         the best terms are kept (see [1]). Note: the method is slow for
%         kappa>3. The suggested value (default) is kappa=2.
%
% the first three must be vectors of the same length, and kappa a scalar
%
% outputs:
%  R    : O(n^-kappa) approximation of the PPV meta-converse bound on rate
%
% The method builds upon the O(n^-k) approximation of the PPV bound from:
% [1] T. Erseghe, "Coding in the Finite-Blocklength Regime: Bounds based 
% on Laplace Integrals and their Asymptotic Approximations", 
% http://arxiv.org/abs/1511.04629
%
% (c) T. Erseghe 2016

if ~exist('kappa')
    kappa = 2; % default value
end

options1 = optimset('Algorithm','interior-point',...
    'GradObj','on',...
    'GradCon','on',...
    'Display', 'off',...
    'TolFun',1e-20,...
    'TolX', 1e-20,...
    'MaxIter',1000);
options2 = optimset('Algorithm','interior-point',...
    'GradObj','on',...
    'GradCon','off',...
    'Display', 'off',...
    'TolFun',1e-20,...
    'TolX', 1e-20,...
    'MaxIter',1000);

for k = 1:length(n)
    if (n(k)==inf)
        Pb = 0.5*erfc(sqrt(Om(k)/2));
        R(k) = 1+Pb.*log(Pb)+(1-Pb).*log(1-Pb);
    else
        Pb = 0.5*erfc(sqrt(Om(k)/2));
        nn = n(k);
        pe = Pe(k);
        % 1) find lambda
        Qe = fmincon(@(x)zero_obj(x),0,[],[],[],[],[],[],...
                     @(x)constr_Qe(x,pe),options1);
        la0 = min(Pb+sqrt(Pb*(1-Pb)/nn)*Qe,0.4999);
        if ~isreal(log_MD(la0, Pb, nn, kappa))
            % not found an initialization that always works...
            % ... left to be solved in future
            R(k) = nan;
        else
            la = fmincon(@(x)zero_obj(x),la0,[],[],[],[],[],[],...
                         @(x)constr_Pe(x,Pb,nn,kappa,pe),options2);
            % 2) evaluate the rate value
            R(k) = -log_FA(la,nn,kappa)/log(2);
            if (R(k)<0)||(~isreal(R(k)))
                R(k) = nan;
            end
        end
    end
end


% ------------------- FUNCTIONS FOR OPTIMIZATION -------------------------

function y = log_FA(la, n, kappa)
if (la>=0.5)
    y = nan;
    return;
end
N = kappa-1;
G = 2;
Gla = 2*la;
al = -la*log(la)-(1-la)*log(1-la)-log(2);
c0 = sqrt(2*(1-la)/la)/(1-la*G);
c2 = -sqrt(2/la/(1-la))*(1-la+la^2-2*G*la*(1+5*la-5*la^2)+...
           G^2*la^2*(13-13*la+la^2))/(6*la*(1-la*G)^3);
if (kappa==2) % O(n^-2) approx in [1]
    ck = [c0];
elseif (kappa==3) % O(n^-3) approx in [1]
    ck = [c0 c2];
else % O(n^-k) approx in [1]
    ck = coefficients(la,G,N,c0,c2);
end
g = (ck/(2*pi)).*gamma(0.5:1:N-0.5)./(n.^(0.5:1:N-0.5));
tg = cumsum(g);
if N>1
    [~,pos] = min(abs(g(2:end)));
    pos = pos(1)+1;
    tg = tg(pos);
end
% pre = abs(g(pos))/abs(tg); % precision
y = al+log(tg)/n;

function y = log_MD(la, Pb, n, kappa)
if (la<=Pb)||(la>=0.5)
    y = nan;
    return;
end
N = kappa-1;
G = 1/Pb;
Gla = la/Pb;
al = -la*log(Gla)-(1-la)*log((1/Pb-Gla)/(1/Pb-1));
c0 = sqrt(2*(1-la)/la)/(1-la*G);
c2 = -sqrt(2/la/(1-la))*(1-la+la^2-2*G*la*(1+5*la-5*la^2)+...
           G^2*la^2*(13-13*la+la^2))/(6*la*(1-la*G)^3);
if (kappa==2) % O(n^-2) approx in [1]
    ck = [c0];
elseif (kappa==3) % O(n^-3) approx in [1]
    ck = [c0 c2];
else % O(n^-k) approx in [1]
    ck = coefficients(la,G,N,c0,c2);
end
g = -(ck/(2*pi)).*gamma(0.5:1:N-0.5)./(n.^(0.5:1:N-0.5));
tg = cumsum(g);
if N>1
    [~,pos] = min(abs(g(2:end)));
    tg = tg(pos(1)+1);
end
% pre = abs(g(pos))/abs(tg); % precision
y = al+log(tg)/n;

function ck = coefficients(la,G,N,c0,c2)
Nl = N;
co = 6*(-1).^(1:Nl)./gamma(6:2:2*Nl+4);
L = ones(1,Nl);
for el = 1:Nl-1
    L(el+1,:) = co(1:el) * (L(end:-1:1,:) .* ((1:el)'*(2:Nl+1)-el)/el);
end
cl = 0;
for k = 1:Nl
    cl(2*k+1) = (-1./((1:k).*6.^(1:k)))*diag(L(k:-1:1,1:k));
end
%%%
Nd = N-1;
cd = 0;
for k = 1:Nd+1
    cd(2*k+1) = cl(2*k+1)*sum((-la).^(1:2*k).*gamma(2*k+2)./gamma(2:2*k+1)./gamma(2*k+1:-1:2));
end
d2 = cd(3);
co = cd(5:2:end)/d2;
cu = [0 sqrt(d2) 0 co(1)*sqrt(d2)/2];
for k = 2:Nd
    cu(2*k+2) = 0.5*(cu(2)*co(k)-sum(co(k-1:-1:1).*cu(4:2:end))/k);
end
%%%
Np = N-1;
co = cu(4:2:end)/cu(2);
P = ones(1,2*Np+1);
for el = 1:Np
    P(el+1,:) = co(1:el) * (P(end:-1:1,:) .* ((1:el)'*(2:2*Np+2)-el)/el);
end
cp = [0 1/cu(2)];
for k=1:Np
    cp(2*k+2) = -(cp(2:2:end).*cu(2).^(-2*k:2:-2))*diag(P(k+1:-1:2,1:2:2*k-1));
end
%%%
Nf = 2*N;
k = 1:2*N+4;
nu = 1i.^k.*((la+1).^k -la.^k -2.^k +1); 
nu = nu.*2.^k./gamma(k+1);
nu = nu(2:end);
de = 1i.^k.*((2.^k-2).*(la+G*la.^k)-(la*G+1).*((la+1).^k-1-la.^k)).*2.^k./gamma(k+1);
de = de(3:end);
cf = nu(1)/de(1);
for m = 1:Nf
    cf(m+1) = (nu(m+1) - sum(cf(end:-1:1).*de(2:m+1)))/de(1);
end
%%%
Nc = N-1;
co = cp(4:2:end)/cp(2);
U = ones(1,2*Nc+1);
P = 1;
for el = 1:Nc
    U(el+1,:) = co(1:el) * (U(end:-1:1,:) .* ((1:el)'*(2:2*Nc+2)-el)/el);
    P(el+1,1) = -co(1:el)*P(end:-1:1,1);
end
ck = 2*cf(1)/1i/cp(2);
for k = 1:Nc
    ck(k+1) = ck(1)*P(k+1)-2i*(cf(2*k+1:-2:3).*cp(2).^(2*k-1:-2:1))*diag(U(1:k,2*k-1:-2:1));
end
% check correctness
ch = [c0 c2];
if max(abs(ch-ck(1:2))./abs(ch))>1e-5
    disp('oops!!!')
end
ck(1:2) = ch;

function [f gradf] = zero_obj(x)
f = [0];
gradf = [0];

function [c ceq gradc gradceq] = constr_Qe(x,Pe)
c = [];
gradc = [];
ceq = [0.5*erfc((x)/sqrt(2))-Pe];
gradceq = [-exp(-x.^2/2)/sqrt(2*pi)];

function [c ceq gradc gradceq] = constr_Pe(x,Pb,n,kappa,Pe)
c = [];
gradc = [];
ceq = [log_MD(x,Pb,n,kappa)-log(Pe)/n];
gradceq = []; % too long to derive !!!

