function R = rate_BSC_RCU_On2_LB(n, Om, Pe)
%
% R = rate_BSC_RCU_On2_LB(n, Om, Pe)
%
% inputs :
%  n    : packet length
%  Om   : SNR value
%  Pe   : reference packet error probability
%
% must be vectors of the same length 
%
% outputs:
%  R    : O(n^-2) approximation of the RCU bound on rate
%
% The method builds upon the O(n^-2) approximation of the RCU bound from:
% T. Erseghe, "Coding in the Finite-Blocklength Regime: Bounds based on 
% Laplace Integrals and their Asymptotic Approximations", 
% http://arxiv.org/abs/1511.04629
%
% (c) T. Erseghe 2016

options1 = optimset('Algorithm','interior-point',...
    'GradObj','on',...
    'GradCon','on',...
    'Display', 'off',...
    'TolFun',1e-20,...
    'TolX', 1e-20,...
    'MaxIter',1000);
options2 = optimset('Algorithm','interior-point',...
    'GradObj','on',...
    'GradCon','off',...
    'Display', 'off',...
    'TolFun',1e-20,...
    'TolX', 1e-20,...
    'MaxIter',1000);

for k = 1:length(n)
    if (n(k)==inf)
        Pb = 0.5*erfc(sqrt(Om(k)/2));
        R(k) = 1+Pb.*log(Pb)+(1-Pb).*log(1-Pb);
    else
        nn = n(k);
        Pb = 0.5*erfc(sqrt(Om(k)/2));
        pe = Pe(k);
        % 1) find lambda
        la0 = min(1.011*Pb,1-1e-8); % this works ok
        la = fmincon(@(x)zero_obj(x),la0,[],[],[],[],[],[],...
                     @(x)constr_Pe(x,Pb,nn,pe),options2);
        % 2) evaluate the rate value
        R(k) = rate(la,nn)/log(2);
        if (R(k)<0)||(~isreal(R(k)))
            R(k) = nan;
        end
    end
end


% ------------------- FUNCTIONS FOR OPTIMIZATION -------------------------

function y = rate(s, n)
he = -s.*log(s)-(1-s).*log(1-s);
w2 = log(2)-he+log(2*pi*n*s.*(1-2*s).^2./(1-s))/2/n;
y = w2;

function y = log_Pe(s, Pb, n)
he = -s.*log(s)-(1-s).*log(1-s);
w0 = he+log(1-Pb)-s*log((1-Pb)/Pb)-log(2*pi*n*s.*(1-s))/2/n;
w1 = Pb*(1-s)./(s-Pb)+Pb*(1-s).^2./(Pb*(1-s).^2-s.^2*(1-Pb));
if(w1>0)
    y = w0+log(w1)/n;
else 
    y = nan;
end
    
function [f gradf] = zero_obj(x)
f = [0];
gradf = [0];

function [c ceq gradc gradceq] = constr_Qe(x,Pe)
c = [];
gradc = [];
ceq = [0.5*erfc((x)/sqrt(2))-Pe];
gradceq = [-exp(-x.^2/2)/sqrt(2*pi)];

function [c ceq gradc gradceq] = constr_Pe(x,Pb,n,Pe)
c = [];
gradc = [];
ceq = [log_Pe(x,Pb,n)-log(Pe)/n];
gradceq = []; % too long to derive !!!
