function [R, info] = waterfilling_NA(n, K, sig2, Pe, Ptot)
%
% [R, info_out] = waterfilling_NA(n, K, sig2, Pe, Ptot)
%
% inputs:
%  n    : packet length (must be a multiple of K)
%  K    : block length
%  Pe   : reference packet error probability
%  Ptot : available power [W]
%  sig2 : noise level vector [W], typically N0*B/|G_ch(f)|^2 
%                                 for K different frequencies
%
% outputs:
%  R    : normal approximation bound on rate
%  info : auxiliary informations
%
% The method used builds upon the normal approximation, e.g., see 
% T. Erseghe, "Coding in the Finite-Blocklength Regime: Bounds based on 
% Laplace Integrals and their Asymptotic Approximations", 
% http://arxiv.org/abs/1511.04629
%
% (c) T. Erseghe 2016

sig2 = reshape(sig2,1,length(sig2)); % force it to be a row vector

q = @(x) 0.5*erfc((x)/sqrt(2));
Qe = fminsearch(@(x) abs(q(x)-Pe),0,optimset('TolX',1e-8,'TolFun',1e-8));

% set fmincon options
options = optimset('Algorithm','interior-point',...
    'GradObj','on',...
    'Display', 'off',...
    'TolFun',1e-20,...
    'TolX', 1e-20,...
    'MaxIter',1000,...
    'Hessian','user-supplied',...
    'HessFcn',@(x,lambda)hessian(x,lambda,n,K));

% define linear constraint matrices/vectors for fmincon
A = [sig2];
B = Ptot;
Aeq = [];
Beq = [];
LB = [zeros(size(sig2'))];
UB = [inf*ones(size(sig2'))];

% equal power starting point
x0 = (Ptot/K)./sig2.'; 

% run fmincon
[x,fval,mflag,output] = fmincon(@(x)objective(x,n,K,Qe),x0,...
           A,B,Aeq,Beq,LB,UB,[[],[]],options);       

if (mflag>0)
    disp(['solution found (' num2str(mflag) ')'])
else
    disp('error: solution NOT found')
end

% outputs
info.R = -log2(exp(1))*fval;
if (~isreal(info.R))||(info.R<0)||(mflag<=0)
    info.R = NaN;
end
R = info.R;
info.x = x;
info.x0 = x0;
info.fval = fval;
info.mflag = mflag;
info.output = output;
info.n = n;
info.sig2 = sig2;
info.pow = sig2.*x.'; % power allocation


% ------------------- FUNCTIONS FOR OPTIMIZATION -------------------------

function [f gradf] = objective(x,n,K,Qe)
C = sum(log(1+x))/2/K;
V = sum(x.*(2+x)./(1+x).^2)/2/K;
if (n==inf)
    f = -C;
    gradf = -1./(1+x)/2/K;
else
    f = -(C-sqrt(V/n)*Qe+log(n)/2/n);
    gradf = -(1./(1+x)+1./(1+x).^3/sqrt(V*n))/2/K;
end

function h = hessian(x, lambda, n, K)
if (n==inf)
    h = diag(1./(1+x).^2)/2/K;
else
    V = sum(x.*(2+x)./(1+x).^2)/2/K;
    h = diag(1./(1+x).^2+3./(1+x).^4/sqrt(V*n)+1./(1+x).^6/sqrt(V*n)/V)/2/K;
end