function [R, info] = waterfilling_PPV_On2(n, K, sig2, Pe, Ptot)
%
% [R, info_out] = waterfilling_PPV_1st_order(n, K, sig2, Pe, Ptot)
%
% inputs:
%  n    : packet length (must be a multiple of K)
%  K    : block length
%  Pe   : reference packet error probability
%  Ptot : available power [W]
%  sig2 : noise level vector [W], typically N0*B/|G_ch(f)|^2 
%                                 for K different frequencies
%
% outputs:
%  R    : PPV meta-converse bound on rate (O(n^-2) approximation)
%  info : auxiliary informations
%
% The method used builds upon the O(n^-2) approximation available in 
% T. Erseghe, "Coding in the Finite-Blocklength Regime: Bounds based on 
% Laplace Integrals and their Asymptotic Approximations", 
% http://arxiv.org/abs/1511.04629
%
% Implementation relies on a gradient + Hessian search
%
% (c) T. Erseghe 2016

sig2 = reshape(sig2,1,length(sig2)); % force it to be a row vector

% -----------------------GENERAL REMARK ----------------------------------
%
% we use vector 
%               x = [la, sa, Om] 
% where
%   la : lambda, the threshold level
%   sa : s_alpha, the saddle point
%   Om : Omega_k, the SNR vector, equal to 'Om=pow./sig2' 
%                 with 'pow' the power allocation vector
%
% we maximize function log2(P_FA(x))/n under the constraints:
%
%  alpha'(sa) = 0 - saddle point definition
%  P_FA(x) = Pe - fixed error probability level
%  
% and the linear constraints
%  la >= 0
%  sa >= 0, sa <= 1/2
%  Om >= 0
%  sum(Om.*sig2) = Ptot or, equivalently, sum(pow) = Ptot
% ------------------------------------------------------------------------

% set fmincon options
options = optimset('Algorithm','interior-point',...
    'GradObj','on',...
    'GradCon','on',...
    'Display', 'off',...
    'TolFun',1e-20,...
    'TolX', 1e-20,...
    'MaxIter',1000,...
    'Hessian','user-supplied',...
    'HessFcn',@(x,lambda)hessian(x,lambda,n,K));

% define linear constraint matrices/vectors for fmincon
A = [0, 0, sig2];
B = Ptot;
Aeq = [];
Beq = [];
LB = [0;0;zeros(size(sig2'))];
UB = [inf;1/2;inf*ones(size(sig2'))];

% equal power starting point
x0 = [1; 0.5; (Ptot/K)./sig2.']; 

% run fmincon
[x,fval,mflag,output] = fmincon(@(x)objective(x,n,K),x0,...
           A,B,Aeq,Beq,LB,UB,@(x)constraints(x,n,K,Pe),options);       

if (mflag>0)
    disp(['solution found (' num2str(mflag) ')'])
else
    disp('error: solution NOT found')
end

% outputs
info.R = -log2(exp(1))*fval;
if (~isreal(info.R))||(info.R<0)||(mflag<=0)
    info.R = NaN;
end
R = info.R;
info.x = x;
info.x0 = x0;
info.fval = fval;
info.mflag = mflag;
info.output = output;
info.n = n;
info.sig2 = sig2;
info.pow = sig2.*x(3:end).'; % power allocation
info.lambda = x(1);
info.sa = x(2);
info.Omega = x(3:end);


% ------------------- FUNCTIONS FOR OPTIMIZATION -------------------------

function y = gmFA (m, sa, Om) 
x = 1+2*sa*Om;
if (m==0) % function alpha(sa)-2*la*sa
    y = -2*sa.*(1+Om)./x-log(x);
else % and its derivatives wrt sa
    y = (-2)^m*Om.^(m-1).*(1+Om+Om.*x/m)./x.^(m+1);
end

function y = dgmFA (m, sa, Om) % 1st derivatives wrt Om
x = 1+2*sa*Om;
if (m==0) 
    y = -4*sa*(1-sa+sa*Om)./x.^2;
elseif (m==1)
    y = -4*(1-2*sa)./x.^3;
elseif (m==2)
    y = 4*(3*(1+Om)-2*x)./x.^4;
else
    y = (-2)^m*Om.^(m-2).*((m+1)*(1+Om)-2*x)./x.^(m+2);
end
    
function y = ddgmFA (m, sa, Om) % 2nd derivatives wrt Om
x = 1+2*sa*Om;
if (m==0) 
    y = 4*sa^2*(2*sa*Om+3-4*sa)./x.^3;
elseif (m==1)
    y = 24*sa*(1-2*sa)./x.^4;
elseif (m==2)
    y = 12*(1-2*sa)./x.^5;
end

function y = gmMD (m, sa, Om)
if (m==0) % function beta(sa)-2*la*sb-other constants
    x = 1+2*sa*Om;
    y = -(2*sa-1)./x-log(x)+log(1+Om);
else % and its derivatives wrt sa
    y = gmFA(m, sa, Om);
end

function y = dgmMD (m, sa, Om) % 1st derivatives wrt Om
if (m==0) 
    sb = sa-1/2;
    x = 1+2*sa*Om;
    y = 4*sb^2./x.^2./(1+Om);
else
    y = dgmFA(m, sa, Om);
end

function y = ddgmMD (m, sa, Om) % 2nd derivatives wrt Om
if (m==0) 
    sb = sa-1/2;
    x = 1+2*sa*Om;
    y = -4*sb^2.*(3*x+4*sb)./x.^3./(1+Om).^2;
else
    y = ddgmFA(m, sa, Om);
end

function [f gradf] = objective(x,n,K)
la = x(1);
sa = x(2);
Om = x(3:end);
% objective value
a0 = sum(gmFA(0,sa,Om)/K)+2*sa*la;
a1 = sum(gmFA(1,sa,Om)/K)+2*la;
a2 = sum(gmFA(2,sa,Om)/K);
a3 = sum(gmFA(3,sa,Om)/K);
a0g = dgmFA(0,sa,Om)/K;
a2g = dgmFA(2,sa,Om)/K;
if (n==inf)
    f = a0/2;
    gradf = [sa; a1/2; a0g/2];
else
    f = a0/2-log(n)/2/n-log(2*pi*a2*sa^2)/2/n;
    gradf = [sa; ...
             a1/2-(1.5*a3/a2+1/sa)/n;...
             a0g/2-(a2g/a2/2)/n];
end

function [c ceq gradc gradceq] = constraints(x,n,K,Pe)
la = x(1);
sa = x(2);
sb = sa - 1/2;
Om = x(3:end);
% inequality constraints (none)
c = [];
gradc = [];
% equality constraints 
a1 = sum(gmFA(1,sa,Om)/K)+2*la;
a2 = sum(gmFA(2,sa,Om)/K);
a1g = dgmFA(1,sa,Om)/K;
b0 = sum(gmMD(0,sa,Om)/K)+2*sb*la;
b1 = sum(gmMD(1,sa,Om)/K)+2*la;
b2 = sum(gmMD(2,sa,Om)/K);
b3 = sum(gmMD(3,sa,Om)/K);
b0g = dgmMD(0,sa,Om)/K;
b2g = dgmMD(2,sa,Om)/K;
if n==inf
    ceq = [a1, b0/2].';
    gradceq = [2, 2*a2, a1g.'; ...
               sb, b1/2, (b0g/2).'].';
else
    ceq = [a1, b0/2-(2*log(Pe)+log(n)+log(2*pi*b2*sb^2))/(2*n)].';
    gradceq = [2, 2*a2, a1g.'; ...
               sb, b1/2-(1.5*b3/b2+1/sb)/n, (b0g/2-b2g/b2/2/n).'].';
end

function h = hessian(x, lambda, n, K)
la = x(1);
sa = x(2);
sb = sa - 1/2;
Om = x(3:end);
a2 = sum(gmFA(2,sa,Om)/K);
a3 = sum(gmFA(3,sa,Om)/K);
a4 = sum(gmFA(4,sa,Om)/K);
a1g = dgmFA(1,sa,Om)/K;
a2g = dgmFA(2,sa,Om)/K;
a3g = dgmFA(3,sa,Om)/K;
a0gg = ddgmFA(0,sa,Om)/K;
a1gg = ddgmFA(1,sa,Om)/K;
a2gg = ddgmFA(2,sa,Om)/K;
b2 = sum(gmMD(2,sa,Om)/K);
b3 = sum(gmMD(3,sa,Om)/K);
b4 = sum(gmMD(4,sa,Om)/K);
b1g = dgmMD(1,sa,Om)/K;
b2g = dgmMD(2,sa,Om)/K;
b3g = dgmMD(3,sa,Om)/K;
b0gg = ddgmMD(0,sa,Om)/K;
b1gg = ddgmMD(1,sa,Om)/K;
b2gg = ddgmMD(2,sa,Om)/K;
zer = zeros(length(Om),1);
% target function
if (n==inf)
    A = a2;
    B = a1g/2;
    C = diag(a0gg/2);
else
    A = a2-(1.5*(4*a4*a2-3*a3^2)/a2^2-1/sa^2)/n;
    B = a1g/2 - 1.5*(a2*a3g-a3*a2g)/a2^2/n;
    C = diag(a0gg/2-a2gg/a2/2/n)+ a2g*a2g.'/a2^2/2/n;
end
h = [  0, 1, zer.'; ...
       1, A,   B.'; ...
     zer, B,   C];
% inequality constraint #1 - alpha'(sa)=0
A = 6*a3;
B = 2*a2g;
C = diag(a1gg);
hessc = [  0, 0, zer.'; ...
           0, A,   B.'; ...
         zer, B,   C];
h = h + hessc*lambda.eqnonlin(1);
% inequality constraint #2 - P_MD(x) = Pe
if (n==inf)
    A = b2;
    B = b1g/2;
    C = diag(b0gg/2);
else
    A = b2-(1.5*(4*b4*b2-3*b3^2)/b2^2-1/sb^2)/n;
    B = b1g/2 - 1.5*(b2*b3g-b3*b2g)/b2^2/n;
    C = diag(b0gg/2-b2gg/b2/2/n)+ b2g*b2g.'/b2^2/2/n;
end
hessc = [  0, 0, zer.'; ...
           0, A,   B.'; ...
         zer, B,   C];    
h = h + hessc*lambda.eqnonlin(2);
